from .pool import NodePool


class PomiceEvent:
    """The base class for all events dispatched by a node. 
       Every event must be formatted within your bot's code as a listener.
       i.e: If you want to listen for when a track starts, the event would be:
       ```py
       @bot.listen
       async def on_pomice_track_start(self, event):
       ```
    """
    name = "event"


class TrackStartEvent(PomiceEvent):
    """Fired when a track has successfully started.
       Returns the player associated with the track and the track ID
    """

    def __init__(self, data):
        super().__init__()

        self.name = "track_start"
        self.player = NodePool.get_node().get_player(int(data["guildId"]))
        self.track_id = data["track"]

    def __repr__(self) -> str:
        return f"<Pomice.TrackStartEvent track_id={self.track_id}>"


class TrackEndEvent(PomiceEvent):
    """Fired when a track has successfully ended.
       Returns the player associated with the track along with the track ID and reason.
    """

    def __init__(self, data):
        super().__init__()

        self.name = "track_end"
        self.player = NodePool.get_node().get_player(int(data["guildId"]))
        self.track_id = data["track"]
        self.reason = data["reason"]

    def __repr__(self) -> str:
        return f"<Pomice.TrackEndEvent track_id={self.track_id} reason={self.reason}>"


class TrackStuckEvent(PomiceEvent):
    """Fired when a track is stuck and cannot be played. Returns the player
       associated with the track along with a track ID to be further parsed by the end user.
    """

    def __init__(self, data):
        super().__init__()

        self.name = "track_stuck"
        self.player = NodePool.get_node().get_player(int(data["guildId"]))

        self.track_id = data["track"]
        self.threshold = data["thresholdMs"]

    def __repr__(self) -> str:
        return f"<Pomice.TrackStuckEvent track_id={self.track_id} threshold={self.threshold}>"


class TrackExceptionEvent(PomiceEvent):
    """Fired when a track error has occured.
       Returns the player associated with the track along with the error code and exception.
    """

    def __init__(self, data):
        super().__init__()

        self.name = "track_exception"
        self.player = NodePool.get_node().get_player(int(data["guildId"]))

        self.error = data["error"]
        self.exception = data["exception"]

    def __repr__(self) -> str:
        return f"<Pomice.TrackExceptionEvent> error={self.error} exeception={self.exception}"


class WebsocketClosedEvent(PomiceEvent):
    """Fired when a websocket connection to a node has been closed.
       Returns the reason and the error code.
    """

    def __init__(self, data):
        super().__init__()

        self.name = "websocket_closed"

        self.reason = data["reason"]
        self.code = data["code"]

    def __repr__(self) -> str:
        return f"<Pomice.WebsocketClosedEvent reason={self.reason} code={self.code}>"


class WebsocketOpenEvent(PomiceEvent):
    """Fired when a websocket connection to a node has been initiated.
       Returns the target and the session SSRC.
    """

    def __init__(self, data):
        super().__init__()

        self.name = "websocket_open"

        self.target: str = data["target"]
        self.ssrc: int = data["ssrc"]

    def __repr__(self) -> str:
        return f"<Pomice.WebsocketOpenEvent target={self.target} ssrc={self.ssrc}>"



from urllib.parse import quote_plus as quote
from typing import Optional, Dict, Iterable, Union, Set, Callable, Tuple, Any

VALID_SCOPES = (
    # Playlists
    "playlist-read-collaborative"
    "playlist-modify-private"
    "playlist-modify-public"
    "playlist-read-private"
    # Spotify Connect
    "user-modify-playback-state"
    "user-read-currently-playing"
    "user-read-playback-state"
    # Users
    "user-read-private"
    "user-read-email"
    # Library
    "user-library-modify"
    "user-library-read"
    # Follow
    "user-follow-modify"
    "user-follow-read"
    # Listening History
    "user-read-recently-played"
    "user-top-read"
    # Playback
    "streaming"
    "app-remote-control"
)


def set_required_scopes(*scopes: Optional[str]) -> Callable:
    """A decorator that lets you attach metadata to functions.

    Parameters
    ----------
    scopes : :class:`str`
        A series of scopes that are required.

    Returns
    -------
    decorator : :class:`typing.Callable`
        The decorator that sets the scope metadata.
    """

    def decorate(func) -> Callable:
        func.__requires_spotify_scopes__ = tuple(scopes)
        return func

    return decorate


def get_required_scopes(func: Callable[..., Any]) -> Tuple[str, ...]:
    """Get the required scopes for a function.

    Parameters
    ----------
    func : Callable[..., Any]
        The function to inspect.

    Returns
    -------
    scopes : Tuple[:class:`str`, ...]
        A tuple of scopes required for a call to succeed.
    """
    if not hasattr(func, "__requires_spotify_scopes__"):
        raise AttributeError("Scope metadata has not been set for this object!")
    return func.__requires_spotify_scopes__  # type: ignore


class OAuth2:
    """Helper object for Spotify OAuth2 operations.

    At a very basic level you can you oauth2 only for authentication.

    >>> oauth2 = OAuth2(client, 'some://redirect/uri')
    >>> print(oauth2.url)

    Working with scopes:

    >>> oauth2 = OAuth2(client, 'some://redirect/uri', scopes=['user-read-currently-playing'])
    >>> oauth2.set_scopes(user_read_playback_state=True)

    Parameters
    ----------
    client_id : :class:`str`
        The client id provided by spotify for the app.
    redirect_uri : :class:`str`
        The URI Spotify should redirect to.
    scopes : Optional[Iterable[:class:`str`], Dict[:class:`str`, :class:`bool`]]
        The scopes to be requested.
    state : Optional[:class:`str`]
        The state used to secure sessions.

    Attributes
    ----------
    attrs : Dict[:class:`str`, :class:`str`]
        The attributes used when constructing url parameters
    parameters : :class:`str`
        The URL parameters used
    url : :class:`str`
        The URL for OAuth2
    """

    _BASE = "https://accounts.spotify.com/authorize/?response_type=code&{parameters}"

    def __init__(
        self,
        client_id: str,
        redirect_uri: str,
        *,
        scopes: Optional[Union[Iterable[str], Dict[str, bool]]] = None,
        state: str = None,
    ):
        self.client_id = client_id
        self.redirect_uri = redirect_uri
        self.state = state
        self.__scopes: Set[str] = set()

        if scopes is not None:
            if not isinstance(scopes, dict) and hasattr(scopes, "__iter__"):
                scopes = {scope: True for scope in scopes}

            if isinstance(scopes, dict):
                self.set_scopes(**scopes)
            else:
                raise TypeError(
                    f"scopes must be an iterable of strings or a dict of string to bools"
                )

    def __repr__(self):
        return f"<spotfy.OAuth2: client_id={self.client_id!r}, scope={self.scopes!r}>"

    def __str__(self):
        return self.url

    # Alternate constructors

    @classmethod
    def from_client(cls, client, *args, **kwargs):
        """Construct a OAuth2 object from a `spotify.Client`."""
        return cls(client.http.client_id, *args, **kwargs)

    @staticmethod
    def url_only(**kwargs) -> str:
        """Construct a OAuth2 URL instead of an OAuth2 object."""
        oauth = OAuth2(**kwargs)
        return oauth.url

    # Properties

    @property
    def scopes(self):
        """:class:`frozenset` - A frozenset of the current scopes"""
        return frozenset(self.__scopes)

    @property
    def attributes(self):
        """Attributes used when constructing url parameters."""
        data = {"client_id": self.client_id, "redirect_uri": quote(self.redirect_uri)}

        if self.scopes:
            data["scope"] = quote(" ".join(self.scopes))

        if self.state is not None:
            data["state"] = self.state

        return data

    attrs = attributes

    @property
    def parameters(self) -> str:
        """:class:`str` - The formatted url parameters that are used."""
        return "&".join("{0}={1}".format(*item) for item in self.attributes.items())

    @property
    def url(self) -> str:
        """:class:`str` - The formatted oauth url used for authorization."""
        return self._BASE.format(parameters=self.parameters)

    # Public api

    def set_scopes(self, **scopes: Dict[str, bool]) -> None:
        r"""Modify the scopes for the current oauth2 object.

        Add or remove certain scopes from this oauth2 instance.
        Since hypens are not allowed, replace the _ with -.

        >>> oauth2.set_scopes(user_read_playback_state=True)

        Parameters
        ----------
        \*\*scopes: Dict[:class:`str`, :class:`bool`]
            The scopes to enable or disable.
        """
        for scope_name, state in scopes.items():
            scope_name = scope_name.replace("_","-")
            if state:
                self.__scopes.add(scope_name)
            else:
                self.__scopes.remove(scope_name)



"""Type annotation aliases for other Spotify models."""

from typing import Union, Sequence

from .base import URIBase

SomeURI = Union[URIBase, str]
SomeURIs = Sequence[Union[URIBase, str]]
OneOrMoreURIs = Union[SomeURI, Sequence[SomeURI]]



__all__ = ("SpotifyException", "HTTPException", "Forbidden", "NotFound")


class SpotifyException(Exception):
    """Base exception class for spotify.py."""


class HTTPException(SpotifyException):
    """A generic exception that's thrown when a HTTP operation fails."""

    def __init__(self, response, message):
        self.response = response
        self.status = response.status
        error = message.get("error")

        if isinstance(error, dict):
            self.text = error.get("message", "")
        else:
            self.text = message.get("error_description", "")

        fmt = "{0.reason} (status code: {0.status})"
        if self.text.strip():
            fmt += ": {1}"

        super().__init__(fmt.format(self.response, self.text))


class Forbidden(HTTPException):
    """An exception that's thrown when status code 403 occurs."""


class NotFound(HTTPException):
    """An exception that's thrown when status code 404 occurs."""


class BearerTokenError(HTTPException):
    """An exception that's thrown when Spotify could not provide a valid Bearer Token"""


class RateLimitedException(Exception):
    """An exception that gets thrown when a rate limit is encountered."""



import time
from typing import Any, Dict, Type, Union

import discord
from discord import VoiceChannel, VoiceProtocol, Guild
from discord.ext import commands

from . import events, filters, NodePool, objects, Node 
from .exceptions import TrackInvalidPosition


class Player(VoiceProtocol):
    """The base player class for Pomice.
       In order to initiate a player, you must pass it in as a cls when you connect to a channel.
       i.e: ```py
       await ctx.author.voice.channel.connect(cls=pomice.Player)
       ```
    """

    def __init__(self, client: Type[Union[discord.Client, commands.Bot, commands.AutoShardedBot]], channel: VoiceChannel):
        super().__init__(client=client, channel=channel)

        self.client = client
        self._bot: Union[discord.Client, commands.Bot, commands.AutoShardedBot] = client
        self.channel = channel
        self._guild: discord.Guild = self.channel.guild

        self._node = NodePool.get_node()
        self._current: objects.Track = None
        self._filter: filters.Filter = None
        self._volume = 100
        self._paused = False
        self._is_connected = False

        self._position = 0
        self._last_position = 0
        self._last_update = 0

        self._voice_state = {}

    def __repr__(self):
        return (
            f"<Pomice.player bot={self.bot} guildId={self.guild.id} "
            f"is_connected={self.is_connected} is_playing={self.is_playing}>"
        )

    @property
    def position(self) -> float:
        """Property which returns the player's position in a track in milliseconds"""

        if not self.is_playing or not self.current:
            return 0

        if self.is_paused:
            return min(self._last_position, self.current.length)

        difference = (time.time() * 1000) - self._last_update
        position = self._last_position + difference

        if position > self.current.length:
            return 0

        return min(position, self.current.length)

    @property
    def is_playing(self) -> bool:
        """Property which returns whether or not the player is actively playing a track."""
        return self._is_connected and self.current is not None

    @property
    def is_connected(self) -> bool:
        """Property which returns whether or not the player is connected"""
        return self._is_connected

    @property
    def is_paused(self) -> bool:
        """Property which returns whether or not the player has a track which is paused or not."""
        return self._is_connected and self._paused

    @property
    def current(self) -> objects.Track:
        """Property which returns the currently playing track"""
        return self._current

    @property
    def node(self) -> Node:
        """Property which returns the node the player is connected to"""
        return self._node

    @property
    def guild(self) -> Guild:
        """Property which returns the guild associated with the player"""
        return self._guild

    @property
    def volume(self) -> int:
        """Property which returns the players current volume"""
        return self._volume


    @property
    def filter(self) -> filters.Filter:
        """Property which returns the currently applied filter, if one is applied"""
        return self._filter

    @property
    def bot(self) -> Type[Union[discord.Client, commands.Bot, commands.AutoShardedBot]]:
        """Property which returns the bot associated with this player instance"""
        return self._bot


    async def _update_state(self, data: dict):
        state: dict = data.get("state")
        self._last_update = time.time() * 1000
        self._is_connected = state.get("connected")
        self._last_position = state.get("position")

    async def _dispatch_voice_update(self, voice_data: Dict[str, Any]):
        if {"sessionId", "event"} != self._voice_state.keys():
            return

        await self._node.send(
            op="voiceUpdate",
            guildId=str(self.guild.id),
            **voice_data
        )

    async def on_voice_server_update(self, data: dict):
        self._voice_state.update({"event": data})
        await self._dispatch_voice_update(self._voice_state)

    async def on_voice_state_update(self, data: dict):
        self._voice_state.update({"sessionId": data.get("session_id")})
        if not (channel_id := data.get("channel_id")):
            self.channel = None
            self._voice_state.clear()
            return

        self.channel = self.guild.get_channel(int(channel_id))
        await self._dispatch_voice_update({**self._voice_state, "event": data})

    async def _dispatch_event(self, data: dict):
        event_type = data.get("type")
        event = getattr(events, event_type, None)
        event = event(data)
        self.bot.dispatch(f"pomice_{event.name}", event)

    async def get_tracks(self, query: str, ctx: commands.Context = None):
        """Fetches tracks from the node's REST api to parse into Lavalink.

        If you passed in Spotify API credentials when you created the node, you can also pass in a Spotify URL of a playlist,
        album or track and it will be parsed accordingly.

        You can also pass in a discord.py Context object to get a
        Context object on any track you search.
        """
        return await self._node.get_tracks(query, ctx)

    async def connect(self, *, timeout: float, reconnect: bool):
        await self.guild.change_voice_state(channel=self.channel)
        self._node._players[self.guild.id] = self
        self._is_connected = True

    async def stop(self):
        """Stops a currently playing track."""
        self._current = None
        await self._node.send(op="stop", guildId=str(self.guild.id))

    async def disconnect(self, *, force: bool = False):
        await self.stop()
        await self.guild.change_voice_state(channel=None)
        self.cleanup()
        self.channel = None
        self._is_connected = False
        del self._node._players[self.guild.id]

    async def destroy(self):
        """Disconnects a player and destroys the player instance."""
        await self.disconnect()
        await self._node.send(op="destroy", guildId=str(self.guild.id))

    async def play(self, track: objects.Track, start_position: int = 0) -> objects.Track:
        """Plays a track. If a Spotify track is passed in, it will be handled accordingly."""
        if track.spotify:
            spotify_track: objects.Track = (await self._node.get_tracks(f"{track.search_type}"))[0]
            track.youtube_result = spotify_track
            await self._node.send(
                op="play",
                guildId=str(self.guild.id),
                track=spotify_track.track_id,
                startTime=start_position,
                endTime=spotify_track.length,
                noReplace=False
            )
        else:
            await self._node.send(
                op="play",
                guildId=str(self.guild.id),
                track=track.track_id,
                startTime=start_position,
                endTime=track.length,
                noReplace=False
            )
        self._current = track
        return self._current

    async def seek(self, position: float) -> float:
        """Seeks to a position in the currently playing track milliseconds"""

        if position < 0 or position > self.current.length:
            raise TrackInvalidPosition(
                f"Seek position must be between 0 and the track length"
            )

        await self._node.send(op="seek", guildId=str(self.guild.id), position=position)
        return self._position

    async def set_pause(self, pause: bool) -> bool:
        """Sets the pause state of the currently playing track."""
        await self._node.send(op="pause", guildId=str(self.guild.id), pause=pause)
        self._paused = pause
        return self._paused

    async def set_volume(self, volume: int) -> int:
        """Sets the volume of the player as an integer. Lavalink accepts an amount from 0 to 500."""
        await self._node.send(op="volume", guildId=str(self.guild.id), volume=volume)
        self._volume = volume
        return self._volume

    async def set_filter(self, filter: filters.Filter) -> filters.Filter:
        """Sets a filter of the player. Takes a pomice.Filter object.
           This will only work if you are using the development version of Lavalink.
        """
        await self._node.send(op="filters", guildId=str(self.guild.id), **filter.payload)
        await self.seek(self.position)
        self._filter = filter
        return filter



import asyncio
from typing import Optional, List, Iterable, NamedTuple, Type, Union, Dict

from .http import HTTPClient
from .utils import to_id
from . import OAuth2, Artist, Album, Track, User, Playlist

__all__ = ("Client", "SearchResults")

_TYPES = {"artist": Artist, "album": Album, "playlist": Playlist, "track": Track}

_SEARCH_TYPES = {"track", "playlist", "artist", "album"}
_SEARCH_TYPE_ERR = (
    'Bad queary type! got "%s" expected any of: track, playlist, artist, album'
)


class SearchResults(NamedTuple):
    """A namedtuple of search results.

    Attributes
    ----------
    artists : List[:class:`Artist`]
        The artists of the search.
    playlists : List[:class:`Playlist`]
        The playlists of the search.
    albums : List[:class:`Album`]
        The albums of the search.
    tracks : List[:class:`Track`]
        The tracks of the search.
    """

    artists: Optional[List[Artist]] = None
    playlists: Optional[List[Playlist]] = None
    albums: Optional[List[Album]] = None
    tracks: Optional[List[Track]] = None


class Client:
    """Represents a Client app on Spotify.

    This class is used to interact with the Spotify API.

    Parameters
    ----------
    client_id : :class:`str`
        The client id provided by spotify for the app.
    client_secret : :class:`str`
        The client secret for the app.
    loop : Optional[:class:`asyncio.AbstractEventLoop`]
        The event loop the client should run on, if no loop is specified `asyncio.get_event_loop()` is called and used instead.

    Attributes
    ----------
    client_id : :class:`str`
        The applications client_id, also aliased as `id`
    http : :class:`HTTPClient`
        The HTTPClient that is being used.
    loop : Optional[:class:`asyncio.AbstractEventLoop`]
        The event loop the client is running on.
    """

    _default_http_client: Type[HTTPClient] = HTTPClient

    def __init__(
        self,
        client_id: str,
        client_secret: str,
        *,
        loop: Optional[asyncio.AbstractEventLoop] = None,
    ) -> None:
        if not isinstance(client_id, str):
            raise TypeError("client_id must be a string.")

        if not isinstance(client_secret, str):
            raise TypeError("client_secret must be a string.")

        if loop is not None and not isinstance(loop, asyncio.AbstractEventLoop):
            raise TypeError(
                "loop argument must be None or an instance of asyncio.AbstractEventLoop."
            )

        self.loop = loop = loop or asyncio.get_event_loop()
        self.http = self._default_http_client(client_id, client_secret, loop=loop)

    def __repr__(self):
        return f"<spotify.Client: {self.http.client_id!r}>"

    async def __aenter__(self) -> "Client":
        return self

    async def __aexit__(self, exc_type, exc_value, traceback) -> None:
        await self.close()

    # Properties

    @property
    def client_id(self) -> str:
        """:class:`str` - The Spotify client ID."""
        return self.http.client_id

    @property
    def id(self):  # pylint: disable=invalid-name
        """:class:`str` - The Spotify client ID."""
        return self.http.client_id

    # Public api

    def oauth2_url(
        self,
        redirect_uri: str,
        scopes: Optional[Union[Iterable[str], Dict[str, bool]]] = None,
        state: Optional[str] = None,
    ) -> str:
        """Generate an oauth2 url for user authentication.

        This is an alias to :meth:`OAuth2.url_only` but the
        difference is that the client id is autmatically
        passed in to the constructor.

        Parameters
        ----------
        redirect_uri : :class:`str`
            Where spotify should redirect the user to after authentication.
        scopes : Optional[Iterable[:class:`str`], Dict[:class:`str`, :class:`bool`]]
            The scopes to be requested.
        state : Optional[:class:`str`]
            Using a state value can increase your assurance that an incoming connection is the result of an
            authentication request.

        Returns
        -------
        url : :class:`str`
            The OAuth2 url.
        """
        return OAuth2.url_only(
            client_id=self.http.client_id,
            redirect_uri=redirect_uri,
            scopes=scopes,
            state=state,
        )

    async def close(self) -> None:
        """Close the underlying HTTP session to Spotify."""
        await self.http.close()

    async def user_from_token(self, token: str) -> User:
        """Create a user session from a token.

        .. note::

            This code is equivelent to `User.from_token(client, token)`

        Parameters
        ----------
        token : :class:`str`
            The token to attatch the user session to.

        Returns
        -------
        user : :class:`spotify.User`
            The user from the ID
        """
        return await User.from_token(self, token)

    async def get_album(self, spotify_id: str, *, market: str = "US") -> Album:
        """Retrive an album with a spotify ID.

        Parameters
        ----------
        spotify_id : :class:`str`
            The ID to search for.
        market : Optional[:class:`str`]
            An ISO 3166-1 alpha-2 country code

        Returns
        -------
        album : :class:`spotify.Album`
            The album from the ID
        """
        data = await self.http.album(to_id(spotify_id), market=market)
        return Album(self, data)

    async def get_artist(self, spotify_id: str) -> Artist:
        """Retrive an artist with a spotify ID.

        Parameters
        ----------
        spotify_id : str
            The ID to search for.

        Returns
        -------
        artist : Artist
            The artist from the ID
        """
        data = await self.http.artist(to_id(spotify_id))
        return Artist(self, data)

    async def get_track(self, spotify_id: str) -> Track:
        """Retrive an track with a spotify ID.

        Parameters
        ----------
        spotify_id : str
            The ID to search for.

        Returns
        -------
        track : Track
            The track from the ID
        """
        data = await self.http.track(to_id(spotify_id))
        return Track(self, data)

    async def get_user(self, spotify_id: str) -> User:
        """Retrive an user with a spotify ID.

        Parameters
        ----------
        spotify_id : str
            The ID to search for.

        Returns
        -------
        user : User
            The user from the ID
        """
        data = await self.http.user(to_id(spotify_id))
        return User(self, data)

    # Get multiple objects

    async def get_albums(self, *ids: str, market: str = "US") -> List[Album]:
        """Retrive multiple albums with a list of spotify IDs.

        Parameters
        ----------
        ids : List[str]
            the ID to look for
        market : Optional[str]
            An ISO 3166-1 alpha-2 country code

        Returns
        -------
        albums : List[Album]
            The albums from the IDs
        """
        data = await self.http.albums(
            ",".join(to_id(_id) for _id in ids), market=market
        )
        return list(Album(self, album) for album in data["albums"])

    async def get_artists(self, *ids: str) -> List[Artist]:
        """Retrive multiple artists with a list of spotify IDs.

        Parameters
        ----------
        ids : List[:class:`str`]
            The IDs to look for.

        Returns
        -------
        artists : List[:class:`Artist`]
            The artists from the IDs
        """
        data = await self.http.artists(",".join(to_id(_id) for _id in ids))
        return list(Artist(self, artist) for artist in data["artists"])

    async def search(  # pylint: disable=invalid-name
        self,
        q: str,
        *,
        types: Iterable[str] = ("track", "playlist", "artist", "album"),
        limit: int = 20,
        offset: int = 0,
        market: str = "US",
        should_include_external: bool = False,
    ) -> SearchResults:
        """Access the spotify search functionality.

        >>> results = client.search('Cadet', types=['artist'])
        >>> for artist in result.get('artists', []):
        ...     if artist.name.lower() == 'cadet':
        ...         print(repr(artist))
        ...         break

        Parameters
        ----------
        q : :class:`str`
            the search query
        types : Optional[Iterable[`:class:`str`]]
            A sequence of search types (can be any of `track`, `playlist`, `artist` or `album`) to refine the search request.
            A `ValueError` may be raised if a search type is found that is not valid.
        limit : Optional[:class:`int`]
            The limit of search results to return when searching.
            Maximum limit is 50, any larger may raise a :class:`HTTPException`
        offset : Optional[:class:`int`]
            The offset from where the api should start from in the search results.
        market : Optional[:class:`str`]
            An ISO 3166-1 alpha-2 country code. Provide this parameter if you want to apply Track Relinking.
        should_include_external : :class:`bool`
            If `True` is specified, the response will include any relevant audio content
            that is hosted externally. By default external content is filtered out from responses.

        Returns
        -------
        results : :class:`SearchResults`
            The results of the search.

        Raises
        ------
        TypeError
            Raised when a parameter with a bad type is passed.
        ValueError
            Raised when a bad search type is passed with the `types` argument.
        """
        if not hasattr(types, "__iter__"):
            raise TypeError("types must be an iterable.")

        types_ = set(types)

        if not types_.issubset(_SEARCH_TYPES):
            raise ValueError(_SEARCH_TYPE_ERR % types_.difference(_SEARCH_TYPES).pop())

        query_type = ",".join(tp.strip() for tp in types)

        include_external: Optional[str]
        if should_include_external:
            include_external = "audio"
        else:
            include_external = None

        data = await self.http.search(
            q=q,
            query_type=query_type,
            market=market,
            limit=limit,
            offset=offset,
            include_external=include_external,
        )

        return SearchResults(
            **{
                key: [_TYPES[obj["type"]](self, obj) for obj in value["items"]]
                for key, value in data.items()
            }
        )



"""Source implementation for a spotify User"""

import functools
from functools import partial
from base64 import b64encode
from typing import (
    Optional,
    Dict,
    Union,
    List,
    Type,
    TypeVar,
    TYPE_CHECKING,
)

from ..utils import to_id
from ..http import HTTPUserClient
from . import (
    AsyncIterable,
    URIBase,
    Image,
    Device,
    Context,
    Player,
    Playlist,
    Track,
    Artist,
    Library,
)

if TYPE_CHECKING:
    import spotify

T = TypeVar("T", Artist, Track)  # pylint: disable=invalid-name


def ensure_http(func):
    func.__ensure_http__ = True
    return func


class User(URIBase, AsyncIterable):  # pylint: disable=too-many-instance-attributes
    """A Spotify User.

    Attributes
    ----------
    id : :class:`str`
        The Spotify user ID for the user.
    uri : :class:`str`
        The Spotify URI for the user.
    url : :class:`str`
        The open.spotify URL.
    href : :class:`str`
        A link to the Web API endpoint for this user.
    display_name : :class:`str`
        The name displayed on the user’s profile.
        `None` if not available.
    followers : :class:`int`
        The total number of followers.
    images : List[:class:`Image`]
        The user’s profile image.
    email : :class:`str`
        The user’s email address, as entered by the user when creating their account.
    country : :class:`str`
        The country of the user, as set in the user’s account profile. An ISO 3166-1 alpha-2 country code.
    birthdate : :class:`str`
        The user’s date-of-birth.
    product : :class:`str`
        The user’s Spotify subscription level: “premium”, “free”, etc.
        (The subscription level “open” can be considered the same as “free”.)
    """

    def __init__(self, client: "spotify.Client", data: dict, **kwargs):
        self.__client = self.client = client

        if "http" not in kwargs:
            self.library = None
            self.http = client.http
        else:
            self.http = kwargs.pop("http")
            self.library = Library(client, self)

        # Public user object attributes
        self.id = data.pop("id")  # pylint: disable=invalid-name
        self.uri = data.pop("uri")
        self.url = data.pop("external_urls").get("spotify", None)
        self.display_name = data.pop("display_name", None)
        self.href = data.pop("href")
        self.followers = data.pop("followers", {}).get("total", None)
        self.images = list(Image(**image) for image in data.pop("images", []))

        # Private user object attributes
        self.email = data.pop("email", None)
        self.country = data.pop("country", None)
        self.birthdate = data.pop("birthdate", None)
        self.product = data.pop("product", None)

        # AsyncIterable attrs
        self.__aiter_klass__ = Playlist
        self.__aiter_fetch__ = partial(
            self.__client.http.get_playlists, self.id, limit=50
        )

    def __repr__(self):
        return f"<spotify.User: {(self.display_name or self.id)!r}>"

    def __getattr__(self, attr):
        value = object.__getattribute__(self, attr)

        if (
            hasattr(value, "__ensure_http__")
            and getattr(self, "http", None) is not None
        ):

            @functools.wraps(value)
            def _raise(*args, **kwargs):
                raise AttributeError(
                    "User has not HTTP presence to perform API requests."
                )

            return _raise
        return value

    async def __aenter__(self) -> "User":
        return self

    async def __aexit__(self, _, __, ___):
        await self.http.close()

    # Internals

    async def _get_top(self, klass: Type[T], kwargs: dict) -> List[T]:
        target = {Artist: "artists", Track: "tracks"}[klass]
        data = {
            key: value
            for key, value in kwargs.items()
            if key in ("limit", "offset", "time_range")
        }

        resp = await self.http.top_artists_or_tracks(target, **data)  # type: ignore

        return [klass(self.__client, item) for item in resp["items"]]

    ### Alternate constructors

    @classmethod
    async def from_code(
        cls, client: "spotify.Client", code: str, *, redirect_uri: str,
    ):
        """Create a :class:`User` object from an authorization code.

        Parameters
        ----------
        client : :class:`spotify.Client`
            The spotify client to associate the user with.
        code : :class:`str`
            The authorization code to use to further authenticate the user.
        redirect_uri : :class:`str`
            The rediriect URI to use in tandem with the authorization code.
        """
        route = ("POST", "https://accounts.spotify.com/api/token")
        payload = {
            "redirect_uri": redirect_uri,
            "grant_type": "authorization_code",
            "code": code,
        }

        client_id = client.http.client_id
        client_secret = client.http.client_secret

        headers = {
            "Authorization": f"Basic {b64encode(':'.join((client_id, client_secret)).encode()).decode()}",
            "Content-Type": "application/x-www-form-urlencoded",
        }

        raw = await client.http.request(route, headers=headers, params=payload)

        token = raw["access_token"]
        refresh_token = raw["refresh_token"]

        return await cls.from_token(client, token, refresh_token)

    @classmethod
    async def from_token(
        cls,
        client: "spotify.Client",
        token: Optional[str],
        refresh_token: Optional[str] = None,
    ):
        """Create a :class:`User` object from an access token.

        Parameters
        ----------
        client : :class:`spotify.Client`
            The spotify client to associate the user with.
        token : :class:`str`
            The access token to use for http requests.
        refresh_token : :class:`str`
            Used to acquire new token when it expires.
        """
        client_id = client.http.client_id
        client_secret = client.http.client_secret
        http = HTTPUserClient(client_id, client_secret, token, refresh_token)
        data = await http.current_user()
        return cls(client, data=data, http=http)

    @classmethod
    async def from_refresh_token(cls, client: "spotify.Client", refresh_token: str):
        """Create a :class:`User` object from a refresh token.
        It will poll the spotify API for a new access token and
        use that to initialize the spotify user.

        Parameters
        ----------
        client : :class:`spotify.Client`
            The spotify client to associate the user with.
        refresh_token: str
            Used to acquire token.
        """
        return await cls.from_token(client, None, refresh_token)

    ### Contextual methods

    @ensure_http
    async def currently_playing(self) -> Dict[str, Union[Track, Context, str]]:
        """Get the users currently playing track.

        Returns
        -------
        context, track : Dict[str, Union[Track, Context, str]]
            A tuple of the context and track.
        """
        data = await self.http.currently_playing()  # type: ignore

        if "item" in data:
            context = data.pop("context", None)

            if context is not None:
                data["context"] = Context(context)
            else:
                data["context"] = None

            data["item"] = Track(self.__client, data.get("item", {}) or {})

        return data

    @ensure_http
    async def get_player(self) -> Player:
        """Get information about the users current playback.

        Returns
        -------
        player : :class:`Player`
            A player object representing the current playback.
        """
        player = Player(self.__client, self, await self.http.current_player())  # type: ignore
        return player

    @ensure_http
    async def get_devices(self) -> List[Device]:
        """Get information about the users avaliable devices.

        Returns
        -------
        devices : List[:class:`Device`]
            The devices the user has available.
        """
        data = await self.http.available_devices()  # type: ignore
        return [Device(item) for item in data["devices"]]

    @ensure_http
    async def recently_played(
        self,
        *,
        limit: int = 20,
        before: Optional[str] = None,
        after: Optional[str] = None,
    ) -> List[Dict[str, Union[Track, Context, str]]]:
        """Get tracks from the current users recently played tracks.

        Returns
        -------
        playlist_history : List[Dict[:class:`str`, Union[Track, Context, :class:`str`]]]
            A list of playlist history object.
            Each object is a dict with a timestamp, track and context field.
        """
        data = await self.http.recently_played(limit=limit, before=before, after=after)  # type: ignore
        client = self.__client

        # List[T] where T: {'track': Track, 'content': Context: 'timestamp': ISO8601}
        return [
            {
                "played_at": track.get("played_at"),
                "context": Context(track.get("context", {}) or {}),
                "track": Track(client, track.get("track", {}) or {}),
            }
            for track in data["items"]
        ]

    ### Playlist track methods

    @ensure_http
    async def add_tracks(self, playlist: Union[str, Playlist], *tracks) -> str:
        """Add one or more tracks to a user’s playlist.

        Parameters
        ----------
        playlist : Union[:class:`str`, Playlist]
            The playlist to modify
        tracks : Sequence[Union[:class:`str`, Track]]
            Tracks to add to the playlist

        Returns
        -------
        snapshot_id : :class:`str`
            The snapshot id of the playlist.
        """
        data = await self.http.add_playlist_tracks(  # type: ignore
            to_id(str(playlist)), tracks=[str(track) for track in tracks]
        )
        return data["snapshot_id"]

    @ensure_http
    async def replace_tracks(self, playlist, *tracks) -> None:
        """Replace all the tracks in a playlist, overwriting its existing tracks.

        This powerful request can be useful for replacing tracks, re-ordering existing tracks, or clearing the playlist.

        Parameters
        ----------
        playlist : Union[:class:`str`, PLaylist]
            The playlist to modify
        tracks : Sequence[Union[:class:`str`, Track]]
            Tracks to place in the playlist
        """
        await self.http.replace_playlist_tracks(  # type: ignore
            to_id(str(playlist)), tracks=",".join(str(track) for track in tracks)
        )

    @ensure_http
    async def remove_tracks(self, playlist, *tracks):
        """Remove one or more tracks from a user’s playlist.

        Parameters
        ----------
        playlist : Union[:class:`str`, Playlist]
            The playlist to modify
        tracks : Sequence[Union[:class:`str`, Track]]
            Tracks to remove from the playlist

        Returns
        -------
        snapshot_id : :class:`str`
            The snapshot id of the playlist.
        """
        data = await self.http.remove_playlist_tracks(  # type: ignore
            to_id(str(playlist)), tracks=(str(track) for track in tracks)
        )
        return data["snapshot_id"]

    @ensure_http
    async def reorder_tracks(
        self, playlist, start, insert_before, length=1, *, snapshot_id=None
    ):
        """Reorder a track or a group of tracks in a playlist.

        Parameters
        ----------
        playlist : Union[:class:`str`, Playlist]
            The playlist to modify
        start : int
            The position of the first track to be reordered.
        insert_before : int
            The position where the tracks should be inserted.
        length : Optional[int]
            The amount of tracks to be reordered. Defaults to 1 if not set.
        snapshot_id : :class:`str`
            The playlist’s snapshot ID against which you want to make the changes.

        Returns
        -------
        snapshot_id : :class:`str`
            The snapshot id of the playlist.
        """
        data = await self.http.reorder_playlists_tracks(  # type: ignore
            to_id(str(playlist)), start, length, insert_before, snapshot_id=snapshot_id
        )
        return data["snapshot_id"]

    ### Playlist methods

    @ensure_http
    async def edit_playlist(
        self, playlist, *, name=None, public=None, collaborative=None, description=None
    ):
        """Change a playlist’s name and public/private, collaborative state and description.

        Parameters
        ----------
        playlist : Union[:class:`str`, Playlist]
            The playlist to modify
        name : Optional[:class:`str`]
            The new name of the playlist.
        public : Optional[bool]
            The public/private status of the playlist.
            `True` for public, `False` for private.
        collaborative : Optional[bool]
            If `True`, the playlist will become collaborative and other users will be able to modify the playlist.
        description : Optional[:class:`str`]
            The new playlist description
        """

        kwargs = {
            "name": name,
            "public": public,
            "collaborative": collaborative,
            "description": description,
        }

        await self.http.change_playlist_details(to_id(str(playlist)), **kwargs)  # type: ignore

    @ensure_http
    async def create_playlist(
        self, name, *, public=True, collaborative=False, description=None
    ):
        """Create a playlist for a Spotify user.

        Parameters
        ----------
        name : :class:`str`
            The name of the playlist.
        public : Optional[bool]
            The public/private status of the playlist.
            `True` for public, `False` for private.
        collaborative : Optional[bool]
            If `True`, the playlist will become collaborative and other users will be able to modify the playlist.
        description : Optional[:class:`str`]
            The playlist description

        Returns
        -------
        playlist : :class:`Playlist`
            The playlist that was created.
        """
        data = {"name": name, "public": public, "collaborative": collaborative}

        if description:
            data["description"] = description

        playlist_data = await self.http.create_playlist(self.id, **data)  # type: ignore
        return Playlist(self.__client, playlist_data, http=self.http)

    @ensure_http
    async def follow_playlist(
        self, playlist: Union[str, Playlist], *, public: bool = True
    ) -> None:
        """follow a playlist

        Parameters
        ----------
        playlist : Union[:class:`str`, Playlist]
            The playlist to modify
        public : Optional[bool]
            The public/private status of the playlist.
            `True` for public, `False` for private.
        """
        await self.http.follow_playlist(to_id(str(playlist)), public=public)  # type: ignore

    @ensure_http
    async def get_playlists(
        self, *, limit: int = 20, offset: int = 0
    ) -> List[Playlist]:
        """get the users playlists from spotify.

        Parameters
        ----------
         limit : Optional[int]
             The limit on how many playlists to retrieve for this user (default is 20).
         offset : Optional[int]
             The offset from where the api should start from in the playlists.

        Returns
        -------
        playlists : List[Playlist]
            A list of the users playlists.
        """
        data = await self.http.get_playlists(self.id, limit=limit, offset=offset)  # type: ignore

        return [
            Playlist(self.__client, playlist_data, http=self.http)
            for playlist_data in data["items"]
        ]

    @ensure_http
    async def get_all_playlists(self) -> List[Playlist]:
        """Get all of the users playlists from spotify.

        Returns
        -------
        playlists : List[:class:`Playlist`]
            A list of the users playlists.
        """
        playlists: List[Playlist] = []
        total = None
        offset = 0

        while True:
            data = await self.http.get_playlists(self.id, limit=50, offset=offset)  # type: ignore

            if total is None:
                total = data["total"]

            offset += 50
            playlists += [
                Playlist(self.__client, playlist_data, http=self.http)
                for playlist_data in data["items"]
            ]

            if len(playlists) >= total:
                break

        return playlists

    @ensure_http
    async def top_artists(self, **data) -> List[Artist]:
        """Get the current user’s top artists based on calculated affinity.

        Parameters
        ----------
        limit : Optional[int]
            The number of entities to return. Default: 20. Minimum: 1. Maximum: 50.
        offset : Optional[int]
            The index of the first entity to return. Default: 0
        time_range : Optional[:class:`str`]
            Over what time frame the affinities are computed. (long_term, short_term, medium_term)

        Returns
        -------
        tracks : List[Artist]
            The top artists for the user.
        """
        return await self._get_top(Artist, data)

    @ensure_http
    async def top_tracks(self, **data) -> List[Track]:
        """Get the current user’s top tracks based on calculated affinity.

        Parameters
        ----------
        limit : Optional[int]
            The number of entities to return. Default: 20. Minimum: 1. Maximum: 50.
        offset : Optional[int]
            The index of the first entity to return. Default: 0
        time_range : Optional[:class:`str`]
            Over what time frame the affinities are computed. (long_term, short_term, medium_term)

        Returns
        -------
        tracks : List[Track]
            The top tracks for the user.
        """
        return await self._get_top(Track, data)



import asyncio
import json
import random
import re
import socket
import time
from typing import Optional, Union
from urllib.parse import quote

import aiohttp
import discord
from discord.ext import commands

from typing import Union, Optional


from . import __version__, spotify, SearchType
from .exceptions import (
    InvalidSpotifyClientAuthorization,
    NodeConnectionFailure,
    NodeCreationError,
    NodeNotAvailable,
    NoNodesAvailable,
    SpotifyAlbumLoadFailed,
    SpotifyPlaylistLoadFailed,
    SpotifyTrackLoadFailed,
    TrackLoadError
)
from .objects import Playlist, Track
from .spotify import SpotifyException
from .utils import ExponentialBackoff, NodeStats

SPOTIFY_URL_REGEX = re.compile(
    r"https?://open.spotify.com/(?P<type>album|playlist|track)/(?P<id>[a-zA-Z0-9]+)"
)

DISCORD_MP3_URL_REGEX = re.compile(
    r"https?://cdn.discordapp.com/attachments/(?P<channel_id>[0-9]+)/(?P<message_id>[0-9]+)/(?P<file>[a-zA-Z0-9_.]+)+"
)


class Node:
    """The base class for a node. 
       This node object represents a Lavalink node. 
       To enable Spotify searching, pass in a proper Spotify Client ID and Spotify Client Secret
    """

    def __init__(
        self,
        pool,
        bot: Union[discord.Client, commands.Bot, commands.AutoShardedBot],
        host: str,
        port: int,
        password: str,
        identifier: str,
        session: Optional[aiohttp.ClientSession],
        spotify_client_id: Optional[str],
        spotify_client_secret: Optional[str],
        
    ):
        self._bot: Union[discord.Client, commands.Bot, commands.AutoShardedBot] = bot
        self._host: str = host
        self._port: int = port
        self._pool: NodePool = pool
        self._password: str = password
        self._identifier: str = identifier

        self._websocket_uri: str = f"ws://{self._host}:{self._port}"
        self._rest_uri: str = f"http://{self._host}:{self._port}"

        self._session: aiohttp.ClientSession = session or aiohttp.ClientSession()
        self._websocket: aiohttp.ClientWebSocketResponse = None
        self._task: asyncio.Task = None

        self._connection_id = None
        self._metadata = None
        self._available = None

        self._headers = {
            "Authorization": self._password,
            "User-Id": str(self._bot.user.id),
            "Client-Name": f"Pomice/{__version__}"
        }

        self._players: dict = {}

        self._spotify_client_id: str = spotify_client_id
        self._spotify_client_secret: str = spotify_client_secret

        if self._spotify_client_id and self._spotify_client_secret:
            self._spotify_client = spotify.Client(
                self._spotify_client_id, self._spotify_client_secret
            )
            self._spotify_http_client = spotify.HTTPClient(
                self._spotify_client_id, self._spotify_client_secret
            )

        self._bot.add_listener(self._update_handler, "on_socket_response")

    def __repr__(self):
        return (
            f"<Pomice.node ws_uri={self._websocket_uri} rest_uri={self._rest_uri} "
            f"player_count={len(self._players)}>"
        )

    @property
    def is_connected(self) -> bool:
        """"Property which returns whether this node is connected or not"""
        return self._websocket is not None and not self._websocket.closed

    @property
    async def latency(self) -> int:
        """Property which returns the latency of the node in milliseconds"""
        start_time = time.time()
        await self.send(op="ping")
        end_time = await self._bot.wait_for("node_ping")
        return (end_time - start_time) * 1000

    @property
    async def stats(self) -> NodeStats:
        """Property which returns the node stats."""
        await self.send(op="get-stats")
        node_stats = await self._bot.wait_for("node_stats")
        return node_stats

    @property
    def players(self) -> dict:
        """Property which returns a dict containing the guild ID and the player object."""
        return self._players

    @property
    def bot(self) -> Union[discord.Client, commands.Bot, commands.AutoShardedBot]:
        """Property which returns the discord.py client linked to this node"""
        return self._bot

    @property
    def player_count(self) -> int:
        """Property which returns how many players are connected to this node"""
        return len(self.players)

    @property
    def pool(self):
        """Property which returns the pool this node is apart of"""
        return self._pool

    async def _update_handler(self, data: dict):
        await self._bot.wait_until_ready()

        if not data:
            return

        if data["t"] == "VOICE_SERVER_UPDATE":
            guild_id = int(data["d"]["guild_id"])
            try:
                player = self._players[guild_id]
                await player.on_voice_server_update(data["d"])
            except KeyError:
                return

        elif data["t"] == "VOICE_STATE_UPDATE":
            if int(data["d"]["user_id"]) != self._bot.user.id:
                return

            guild_id = int(data["d"]["guild_id"])
            try:
                player = self._players[guild_id]
                await player.on_voice_state_update(data["d"])
            except KeyError:
                return

    async def _listen(self):
        backoff = ExponentialBackoff(base=7)

        while True:
            msg = await self._websocket.receive()
            if msg.type == aiohttp.WSMsgType.CLOSED:
                retry = backoff.delay()
                await asyncio.sleep(retry)

                if not self.is_connected:
                    self._bot.loop.create_task(self.connect())
            else:
                self._bot.loop.create_task(self._handle_payload(msg.json()))

    async def _handle_payload(self, data: dict):
        op = data.get("op", None)
        if not op:
            return

        if op == "stats":
            self._stats = NodeStats(data)
            return

        if not (player := self._players.get(int(data["guildId"]))):
            return

        if op == "event":
            await player._dispatch_event(data)
        elif op == "playerUpdate":
            await player._update_state(data)

    async def send(self, **data):
        if not self._available:
            raise NodeNotAvailable(
                f"The node '{self.identifier}' is not currently available.")

        await self._websocket.send_str(json.dumps(data))

    def get_player(self, guild_id: int):
        """Takes a guild ID as a parameter. Returns a pomice Player object."""
        return self._players.get(guild_id, None)

    async def connect(self):
        """Initiates a connection with a Lavalink node and adds it to the node pool."""
        await self._bot.wait_until_ready()

        try:
            self._websocket = await self._session.ws_connect(
                self._websocket_uri, headers=self._headers, heartbeat=60
            )
            self._task = self._bot.loop.create_task(self._listen())
            self._available = True
            return self
        except aiohttp.WSServerHandshakeError:
            raise NodeConnectionFailure(
                f"The password for node '{self.identifier}' is invalid."
            )
        except aiohttp.InvalidURL:
            raise NodeConnectionFailure(
                f"The URI for node '{self.identifier}' is invalid."
            )
        except socket.gaierror:
            raise NodeConnectionFailure(
                f"The node '{self.identifier}' failed to connect."
            )

    async def disconnect(self):
        """Disconnects a connected Lavalink node and removes it from the node pool.
           This also destroys any players connected to the node.
        """
        for player in self.players.copy().values():
            await player.destroy()

        await self._websocket.close()
        del self._pool.nodes[self._identifier]
        self.available = False
        self._task.cancel()

    async def get_tracks(self, query: str, ctx: commands.Context = None, search_type: SearchType = None):
        """Fetches tracks from the node's REST api to parse into Lavalink.

           If you passed in Spotify API credentials, you can also pass in a
           Spotify URL of a playlist, album or track and it will be parsed accordingly.

           You can also pass in a discord.py Context object to get a
           Context object on any track you search.
        """
        if spotify_url_check := SPOTIFY_URL_REGEX.match(query):
            if not self._spotify_client_id and not self._spotify_client_secret:
                raise InvalidSpotifyClientAuthorization(
                    "You did not provide proper Spotify client authorization credentials. "
                    "If you would like to use the Spotify searching feature, "
                    "please obtain Spotify API credentials here: https://developer.spotify.com/"
                )

            spotify_search_type = spotify_url_check.group("type")
            spotify_id = spotify_url_check.group("id")

            if spotify_search_type == "playlist":
                results = spotify.Playlist(
                    client=self._spotify_client,
                    data=await self._spotify_http_client.get_playlist(spotify_id)
                )

                try:
                    search_tracks = await results.get_all_tracks()
                    tracks = [
                        Track(
                            track_id=track.id,
                            ctx=ctx,
                            search_type=f"{search_type}{track.artists[0].name} - {track.name}" if search_type else f"ytmsearch:{track.artists[0].name} - {track.name}",
                            spotify=True,
                            info={
                                "title": track.name or "Unknown",
                                "author": ", ".join(
                                    artist.name for artist in track.artists
                                ) or "Unknown",
                                "length": track.duration or 0,
                                "identifier": track.id or "Unknown",
                                "uri": track.url or "spotify",
                                "isStream": False,
                                "isSeekable": False,
                                "position": 0,
                                "thumbnail": track.images[0].url if track.images else None
                            },
                        ) for track in search_tracks
                    ]

                    return Playlist(
                        playlist_info={"name": results.name, "selectedTrack": tracks[0]},
                        tracks=tracks,
                        ctx=ctx,
                        spotify=True,
                        thumbnail=results.images[0].url,
                        uri=results.url,
                    )

                except SpotifyException:
                    raise SpotifyPlaylistLoadFailed(
                        f"Unable to find results for {query}"
                    )

            elif spotify_search_type == "album":
                results = await self._spotify_client.get_album(spotify_id=spotify_id)

                try:
                    search_tracks = await results.get_all_tracks()
                    tracks = [
                        Track(
                            track_id=track.id,
                            ctx=ctx,
                            search_type=f"{search_type}{track.artists[0].name} - {track.name}" if search_type else f"ytmsearch:{track.artists[0].name} - {track.name}",
                            spotify=True,
                            info={
                                "title": track.name or "Unknown",
                                "author": ", ".join(
                                    artist.name for artist in track.artists
                                ) or "Unknown",
                                "length": track.duration or 0,
                                "identifier": track.id or "Unknown",
                                "uri": track.url or "spotify",
                                "isStream": False,
                                "isSeekable": False,
                                "position": 0,
                                "thumbnail": track.images[0].url if track.images else None
                            },
                        ) for track in search_tracks
                    ]

                    return Playlist(
                        playlist_info={"name": results.name, "selectedTrack": tracks[0]},
                        tracks=tracks,
                        ctx=ctx,
                        spotify=True,
                        thumbnail=results.images[0].url,
                        uri=results.url,
                    )

                except SpotifyException:
                    raise SpotifyAlbumLoadFailed(f"Unable to find results for {query}")

            elif spotify_search_type == 'track':
                try:
                    results = await self._spotify_client.get_track(spotify_id=spotify_id)

                    return [
                        Track(
                            track_id=results.id,
                            ctx=ctx,
                            search_type=f"{search_type}{results.artists[0].name} - {results.name}" if search_type else f"ytmsearch:{results.artists[0].name} - {results.name}",
                            spotify=True,
                            info={
                                "title": results.name or "Unknown",
                                "author": ", ".join(
                                    artist.name for artist in results.artists
                                ) or "Unknown",
                                "length": results.duration or 0,
                                "identifier": results.id or "Unknown",
                                "uri": results.url or "spotify",
                                "isStream": False,
                                "isSeekable": False,
                                "position": 0,
                                "thumbnail": results.images[0].url if results.images else None
                            },
                        )
                    ]

                except SpotifyException:
                    raise SpotifyTrackLoadFailed(f"Unable to find results for {query}")

        elif discord_url := DISCORD_MP3_URL_REGEX.match(query):
            async with self._session.get(
                url=f"{self._rest_uri}/loadtracks?identifier={quote(query)}",
                headers={"Authorization": self._password}
            ) as response:
                data: dict = await response.json()


            track: dict = data["tracks"][0]
            info: dict = track.get('info')

            return [Track(
                track_id=track['track'], 
                info={
                    "title": discord_url.group('file'),
                    "author": "Unknown",
                    "length": info.get('length'),
                    "uri": info.get('uri'),
                    "position": info.get('position'),
                    "identifier": info.get('identifier')
                },
                ctx=ctx)]

        else:
            async with self._session.get(
                url=f"{self._rest_uri}/loadtracks?identifier={quote(query)}",
                headers={"Authorization": self._password}
            ) as response:
                data = await response.json()

        load_type = data.get("loadType")

        if not load_type:
            raise TrackLoadError("There was an error while trying to load this track.")

        elif load_type == "LOAD_FAILED":
            raise TrackLoadError(
                f"There was an error of severity '{data['severity']}' "
                f"while loading tracks.\n\n{data['cause']}"
            )

        elif load_type == "NO_MATCHES":
            return None

        elif load_type == "PLAYLIST_LOADED":
            return Playlist(
                playlist_info=data["playlistInfo"],
                tracks=data["tracks"],
                ctx=ctx
            )

        elif load_type == "SEARCH_RESULT" or load_type == "TRACK_LOADED":
            return [
                Track(
                    track_id=track["track"],
                    info=track["info"],
                    ctx=ctx
                )
                for track in data["tracks"]
            ]



class NodePool:
    """The base class for the node pool.
       This holds all the nodes that are to be used by the bot.
    """

    _nodes: dict = {}

    def __repr__(self):
        return f"<Pomice.NodePool node_count={self.node_count}>"

    @property
    def nodes(self) -> dict:
        """Property which returns a dict with the node identifier and the Node object."""
        return self._nodes

    @property
    def node_count(self):
        return len(self._nodes.values())

    @classmethod
    def get_node(cls, *, identifier: str = None) -> Node:
        """Fetches a node from the node pool using it's identifier.
           If no identifier is provided, it will choose a node at random.
        """
        available_nodes = {identifier: node for identifier, node in cls._nodes.items()}
        if not available_nodes:
            raise NoNodesAvailable('There are no nodes available.')

        if identifier is None:
            return random.choice(list(available_nodes.values()))

        return available_nodes.get(identifier, None)

    @classmethod
    async def create_node(
        cls, 
        bot: Union[discord.Client, commands.Bot, commands.AutoShardedBot],
        host: str,
        port: str,
        password: str,
        identifier: str,
        spotify_client_id: Optional[str],
        spotify_client_secret: Optional[str],
        session: Optional[aiohttp.ClientSession] = None,
        
    ) -> Node:
        """Creates a Node object to be then added into the node pool.
           For Spotify searching capabilites, pass in valid Spotify API credentials.
        """
        if identifier in cls._nodes.keys():
            raise NodeCreationError(f"A node with identifier '{identifier}' already exists.")

        node = Node(
            pool=cls, bot=bot, host=host, port=port, password=password,
            identifier=identifier, spotify_client_id=spotify_client_id,
            session=session, spotify_client_secret=spotify_client_secret
        )

        await node.connect()
        cls._nodes[node._identifier] = node
        return node



"""Source implementation for spotify Tracks, and any other semantically relevent, implementation."""

import datetime
from itertools import starmap

from ..oauth import set_required_scopes
from . import URIBase, Image, Artist


class Track(URIBase):  # pylint: disable=too-many-instance-attributes
    """A Spotify Track object.

    Attribtues
    ----------
    id : :class:`str`
        The Spotify ID for the track.
    name : :class:`str`
        The name of the track.
    href : :class:`str`
        A link to the Web API endpoint providing full details of the track.
    uri : :class:`str`
        The Spotify URI for the track.
    duration : int
        The track length in milliseconds.
    explicit : bool
        Whether or not the track has explicit
        `True` if it does.
        `False` if it does not (or unknown)
    disc_number : int
        The disc number (usually 1 unless the album consists of more than one disc).
    track_number : int
        The number of the track.
        If an album has several discs, the track number is the number on the specified disc.
    url : :class:`str`
        The open.spotify URL for this Track
    is_local : bool
        Whether or not the track is from a local file.
    popularity : int
        POPULARITY
    preview_url : :class:`str`
        The preview URL for this Track.
    images : List[Image]
        The images of the Track.
    markets : List[:class:`str`]
        The available markets for the Track.
    """

    def __init__(self, client, data, album=None):
        from .album import Album

        self.__client = client

        self.artists = artists = list(
            Artist(client, artist) for artist in data.pop("artists", [])
        )
        self.artist = artists[-1] if artists else None

        album_ = data.pop("album", None)
        self.album = Album(client, album_) if album_ else album

        self.id = data.pop("id", None)  # pylint: disable=invalid-name
        self.name = data.pop("name", None)
        self.href = data.pop("href", None)
        self.uri = data.pop("uri", None)
        self.duration = data.pop("duration_ms", None)
        self.explicit = data.pop("explicit", None)
        self.disc_number = data.pop("disc_number", None)
        self.track_number = data.pop("track_number", None)
        self.url = data.pop("external_urls").get("spotify", None)
        self.is_local = data.pop("is_local", None)
        self.popularity = data.pop("popularity", None)
        self.preview_url = data.pop("preview_url", None)
        self.markets = data.pop("available_markets", [])

        if "images" in data:
            self.images = list(starmap(Image, data.pop("images")))
        else:
            self.images = self.album.images.copy() if self.album is not None else []

    def __repr__(self):
        return f"<spotify.Track: {self.name!r}>"

    @set_required_scopes(None)
    def audio_analysis(self):
        """Get a detailed audio analysis for the track."""
        return self.__client.http.track_audio_analysis(self.id)

    @set_required_scopes(None)
    def audio_features(self):
        """Get audio feature information for the track."""
        return self.__client.http.track_audio_features(self.id)


class PlaylistTrack(Track, URIBase):
    """A Track on a Playlist.

    Like a regular :class:`Track` but has some additional attributes.

    Attributes
    ----------
    added_by : :class:`str`
        The Spotify user who added the track.
    is_local : bool
        Whether this track is a local file or not.
    added_at : datetime.datetime
        The datetime of when the track was added to the playlist.
    """

    __slots__ = ("added_at", "added_by", "is_local")

    def __init__(self, client, data):
        from .user import User

        super().__init__(client, data["track"])

        self.added_by = User(client, data["added_by"])
        self.added_at = datetime.datetime.strptime(
            data["added_at"], "%Y-%m-%dT%H:%M:%SZ"
        )
        self.is_local = data["is_local"]

    def __repr__(self):
        return f"<spotify.PlaylistTrack: {self.name!r}>"



from re import compile as re_compile
from functools import lru_cache
from contextlib import contextmanager
from typing import Iterable, Hashable, TypeVar, Dict, Tuple

__all__ = ("clean", "filter_items", "to_id")

_URI_RE = re_compile(r"^.*:([a-zA-Z0-9]+)$")
_OPEN_RE = re_compile(r"http[s]?:\/\/open\.spotify\.com\/(.*)\/(.*)")


@contextmanager
def clean(mapping: dict, *keys: Iterable[Hashable]):
    """A helper context manager that defers mutating a mapping."""
    yield
    for key in keys:
        mapping.pop(key)


K = TypeVar("K")  # pylint: disable=invalid-name
V = TypeVar("V")  # pylint: disable=invalid-name


@lru_cache(maxsize=1024)
def _cached_filter_items(data: Tuple[Tuple[K, V], ...]) -> Dict[K, V]:
    data_ = {}
    for key, value in data:
        if value is not None:
            data_[key] = value
    return data_


def filter_items(data: Dict[K, V]) -> Dict[K, V]:
    """Filter the items of a dict where the value is not None."""
    return _cached_filter_items((*data.items(),))


def to_id(value: str) -> str:
    """Get a spotify ID from a URI or open.spotify URL.

    Paramters
    ---------
    value : :class:`str`
        The value to operate on.

    Returns
    -------
    id : :class:`str`
        The Spotify ID from the value.
    """
    value = value.strip()
    match = _URI_RE.match(value)

    if match is None:
        match = _OPEN_RE.match(value)

        if match is None:
            return value
        return match.group(2)
    return match.group(1)
